import { UUID } from '@/types'
import { Dataset, DatasetElementList } from '@/types/dataset'
import { defineStore } from 'pinia'
import { useNotificationStore } from './notification'
import {
  CreateDatasetElementsSelectionParameters,
  CursorPaginationParameters,
  DatasetCreate,
  DatasetEdit,
  cloneDataset,
  createDataset,
  createDatasetElementsSelection,
  deleteDataset,
  deleteDatasetElement,
  listCorpusDataset,
  listDatasetElements,
  retrieveDataset,
  updateDataset
} from '@/api'
import { errorParser } from '@/helpers'

interface State {
  /**
   * Dataset details mapped by ID.
   */
  datasets: { [datasetId: UUID]: Dataset },
  /**
   * Set of dataset IDs per corpora.
   */
  corpusDatasets: { [corpusId: UUID]: Set<UUID> },
  /**
   * List of elements per dataset.
   */
  datasetElements: { [datasetId: UUID]: DatasetElementList[] },
}

export const useDatasetStore = defineStore('dataset', {
  state: (): State => ({
    datasets: {},
    corpusDatasets: {},
    datasetElements: {}
  }),
  actions: {
    async listCorpusDatasets (corpusId: UUID, page = 1) {
      // Do not start fetching corpus datasets if they have been retrieved already
      if (page === 1 && this.corpusDatasets[corpusId]) return
      const data = await listCorpusDataset(corpusId, { page })

      // Fill corpusDatasets[corpusId] with dataset UUIDs
      this.corpusDatasets[corpusId] = new Set([
        ...(this.corpusDatasets[corpusId] ?? []),
        ...data.results.map(dataset => dataset.id)
      ])

      // Add retrieved datasets to this.datasets
      data.results.forEach(newDataset => {
        this.datasets[newDataset.id] = { ...(this.datasets[newDataset.id] ?? {}), ...newDataset }
      })

      if (!data || !data.number || page !== data.number) {
        // Avoid any loop
        throw new Error(`Pagination failed listing datasets for corpus "${corpusId}"`)
      }
      // Load other pages
      if (data.next) await this.listCorpusDatasets(corpusId, page + 1)
    },

    async createCorpusDataset (corpusId: UUID, dataset: DatasetCreate) {
      try {
        const data = await createDataset(corpusId, dataset)
        this.corpusDatasets[corpusId].add(data.id)
        this.datasets[data.id] = data
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async updateCorpusDataset ({ id, ...dataset }: DatasetEdit) {
      try {
        const data = await updateDataset({ id, ...dataset })
        this.datasets[data.id] = data
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async retrieveDataset (datasetId: UUID) {
      try {
        this.datasets[datasetId] = await retrieveDataset(datasetId)
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async deleteCorpusDataset (corpusId: UUID, datasetId: UUID) {
      try {
        await deleteDataset(datasetId)
        if (!this.corpusDatasets[corpusId]) throw new Error(`Datasets for corpus ${corpusId} not found`)
        // Remove datasetId from this.corpusDatasets
        const deleted = this.corpusDatasets[corpusId].delete(datasetId)
        if (!deleted) throw new Error(`Dataset ${datasetId} not found in corpus ${corpusId}`)
        // Remove dataset from this.datasets
        delete this.datasets[datasetId]
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async cloneDataset (datasetId: UUID) {
      try {
        const data = await cloneDataset(datasetId)
        this.datasets[data.id] = data
        if (this.corpusDatasets[data.corpus_id]) this.corpusDatasets[data.corpus_id].add(data.id)
        return data
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    // Add selected elements to a dataset of the same corpus
    async addDatasetElementsSelection (corpusId: UUID, params: CreateDatasetElementsSelectionParameters) {
      try {
        await createDatasetElementsSelection(corpusId, params)
        useNotificationStore().notify({ type: 'success', text: 'Elements have been added to the dataset.' })
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async listDatasetElements (datasetId: UUID, params: CursorPaginationParameters = {}) {
      try {
        const cursor = params.cursor
        const resp = await listDatasetElements(datasetId, { cursor })
        if (!resp.previous) this.datasetElements[datasetId] = []
        this.datasetElements[datasetId] = [
          ...this.datasetElements[datasetId],
          ...resp.results
        ]
        if (resp.next) {
          const cursor = new URL(resp.next).searchParams.get('cursor')
          this.listDatasetElements(datasetId, { cursor })
        }
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    },

    async removeDatasetElement (datasetId: UUID, elementId: UUID, set: string) {
      try {
        await deleteDatasetElement(datasetId, elementId, set)
        const index = this.datasetElements[datasetId].findIndex(datasetElement => datasetElement.set === set && datasetElement.element.id === elementId)
        if (index < 0) throw new Error(`Element ${elementId} not found in set ${set} of dataset ${datasetId}`)
        this.datasetElements[datasetId].splice(index, 1)
      } catch (err) {
        useNotificationStore().notify({ type: 'error', text: errorParser(err) })
        throw err
      }
    }
  },

  getters: {
    /**
     * Retrieve the Dataset objects belonging to a corpus from the list of UUIDs in corpusDatasets
     */
    singleCorpusDatasets () {
      return (corpusId: UUID) => {
        return [...(this.corpusDatasets[corpusId] ?? [])].map(id => this.datasets[id])
      }
    },
    /**
     * Retrieve the elements from one of the Dataset's sets
     */
    datasetSetElements () {
      return (datasetId: UUID, set: string) => (
        this.datasetElements[datasetId]
          ?.filter(datasetElement => datasetElement.set === set)
          .map(({ element }) => element)
      )
    }
  }
})
