import axios from 'axios'
import { assert } from 'chai'
import { pick } from 'lodash'
import { createPinia, setActivePinia } from 'pinia'
import { useModelStore, useNotificationStore } from '@/stores'
import { FakeAxios } from '../testhelpers'

describe('model', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      store = useModelStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('createModel', () => {
      it('creates a model', async () => {
        store.models = { model_0: { id: 'model_0' } }
        const attrs = { name: 'test', description: 'A' }
        mock.onPost('/models/').reply(201, { id: 'model_1', ...attrs })

        await store.createModel(attrs)

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'post',
            url: '/models/',
            data: attrs
          }
        ])
        assert.deepStrictEqual(store.models, {
          model_0: { id: 'model_0' },
          model_1: { id: 'model_1', ...attrs }
        })
      })
    })

    describe('listModels', () => {
      it('lists stored models with filters', async () => {
        const reply = {
          count: 1,
          results: [
            { id: 'model1' }
          ]
        }
        mock.onGet('/models/').reply(200, reply)
        await store.listModels({ name: '1' })
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/models/',
            params: { name: '1' }
          }
        ])
        assert.deepStrictEqual(store.models, { model1: { id: 'model1' } })
      })
    })

    describe('retrieveModel', () => {
      it('retrieves a model', async () => {
        const model = { id: 'model1' }
        mock.onGet('/model/model1/').reply(200, model)

        await store.retrieveModel('model1')

        assert.deepStrictEqual(store.models, {
          model1: model
        })
      })
    })

    describe('listModelVersions', () => {
      it('lists stored model versions', async () => {
        const reply = {
          count: 1,
          results: [
            { id: 'model_version1' }
          ]
        }
        mock.onGet('/model/model1/versions/').reply(200, reply)
        await store.listModelVersions('model1', 1)
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'params'])), [
          {
            method: 'get',
            url: '/model/model1/versions/',
            params: { page: 1 }
          }])
        assert.deepStrictEqual(store.modelVersions, { model_version1: { id: 'model_version1' } })
      })
    })

    describe('getModelVersion', () => {
      it('retrieves a model version', async () => {
        const version = { id: 'model_version1' }
        mock.onGet('/modelversion/model_version1/').reply(200, version)

        await store.getModelVersion('model_version1')

        assert.deepStrictEqual(store.modelVersions, {
          model_version1: version
        })
      })
    })

    describe('deleteModelVersions', () => {
      it('deletes a model version', async () => {
        store.modelVersions = { model_version_0: { id: 'model_version_0' } }

        mock.onDelete('/modelversion/model_version_0/').reply(204)
        await store.deleteModelVersion('model_version_0')
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/modelversion/model_version_0/'
          }])
        assert.deepStrictEqual(store.modelVersions, {})
      })

      it('handles errors', async () => {
        store.modelVersions = { model_version_0: { id: 'model_version_0' } }
        mock.onDelete('/modelversion/model_version_0/').reply(500)

        await store.deleteModelVersion('model_version_0')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/modelversion/model_version_0/'
          }])
        assert.deepStrictEqual(store.modelVersions, { model_version_0: { id: 'model_version_0' } })
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 500'
          }
        ])
      })
    })
  })
})
