import axios from 'axios'
import { assert } from 'chai'
import { pick } from 'lodash'
import { createPinia, setActivePinia } from 'pinia'
import { useNotificationStore, useOAuthStore } from '@/stores'
import { credentialsSample, providersSample } from '../samples'
import { FakeAxios } from '../testhelpers'

describe('oauth', () => {
  describe('actions', () => {
    let mock, store, notificationStore

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      store = useOAuthStore()
      notificationStore = useNotificationStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
      notificationStore.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('listProviders', () => {
      it('lists OAuth providers', async () => {
        mock.onGet('/oauth/providers/').reply(200, providersSample)

        await store.listProviders()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/oauth/providers/'
          }
        ])

        assert.deepStrictEqual(store.providers, providersSample)
        assert.isNull(store.credentials)
      })

      it('handles errors', async () => {
        mock.onGet('/oauth/providers/').reply(418)

        await store.listProviders()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/oauth/providers/'
          }
        ])

        assert.isNull(store.providers)
        assert.isNull(store.credentials)
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 418'
          }
        ])
      })
    })

    describe('listCredentials', () => {
      it('lists OAuth credentials', async () => {
        mock.onGet('/oauth/credentials/').reply(200, credentialsSample)

        await store.listCredentials()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/oauth/credentials/'
          }
        ])

        assert.isNull(store.providers)
        assert.deepStrictEqual(store.credentials, credentialsSample)
      })

      it('handles errors', async () => {
        mock.onGet('/oauth/credentials/').reply(418)

        await store.listCredentials()

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'get',
            url: '/oauth/credentials/'
          }
        ])

        assert.isNull(store.providers)
        assert.isNull(store.credentials)
        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 418'
          }
        ])
      })
    })

    describe('deleteCredentials', () => {
      it('deletes OAuth credentials', async () => {
        mock.onDelete('/oauth/credentials/credid/').reply(204)

        await store.deleteCredentials('credid')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/oauth/credentials/credid/'
          }
        ])
      })

      it('handles errors', async () => {
        mock.onDelete('/oauth/credentials/credid/').reply(418)

        await store.deleteCredentials('credid')

        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url'])), [
          {
            method: 'delete',
            url: '/oauth/credentials/credid/'
          }
        ])

        assert.deepStrictEqual(notificationStore.notifications, [
          {
            id: 0,
            type: 'error',
            text: 'Request failed with status code 418'
          }
        ])
      })
    })
  })
})
